use std::{path::PathBuf, sync::Arc};

use git2::{Cred, PushOptions, RemoteCallbacks, Repository};
use tokio::sync::Mutex;

use crate::{schema::models::GitPushBranch, storage::DynStorageEngine};

use super::GitProvider;

pub struct LocalGitProvider {
    storage_engine: DynStorageEngine,
}

impl LocalGitProvider {
    pub fn new(storage_engine: DynStorageEngine) -> Self {
        Self { storage_engine }
    }
}

#[async_trait::async_trait]
impl GitProvider for LocalGitProvider {
    async fn clone_from_url(&self, url: String) -> eyre::Result<(PathBuf, Repository)> {
        tracing::debug!(url, "allocating dir");
        let dir = self.storage_engine.allocate_dir().await?;

        let dirpath = dir.clone().path();
        let repo = tokio::task::spawn_blocking(move || {
            let mut callbacks = RemoteCallbacks::new();
            callbacks.credentials(|url, username_from_url, _allowed_types| {
                tracing::debug!(username_from_url, url, "pulling key from ssh-agent");
                Cred::ssh_key_from_agent(username_from_url.unwrap())
            });

            let mut fo = git2::FetchOptions::new();
            fo.remote_callbacks(callbacks);

            let mut builder = git2::build::RepoBuilder::new();
            builder.fetch_options(fo);

            tracing::debug!(
                url,
                path = dirpath.as_os_str().to_string_lossy().to_string(),
                "clone git repo"
            );
            builder.clone(url.as_str(), dirpath.as_path())
        })
        .await??;

        tracing::debug!("done pulling repo");

        Ok((dir.path(), repo))
    }

    async fn create_branch(
        &self,
        repo: Arc<Mutex<Repository>>,
        branch: &GitPushBranch,
    ) -> eyre::Result<()> {
        let repo = repo.lock().await;

        let head_commit_oid = repo
            .head()?
            .target()
            .ok_or(eyre::anyhow!("could not get access to target commit"))?;
        let head_commit = repo.find_commit(head_commit_oid)?;
        let newbranch = repo.branch(
            &branch.name.to_lowercase().replace(" ", "-"),
            &head_commit,
            true,
        )?;

        repo.set_head(
            newbranch
                .into_reference()
                .name()
                .ok_or(eyre::anyhow!("could not get name of reference"))?,
        )?;

        Ok(())
    }

    async fn push_branch(
        &self,
        repo: Arc<Mutex<Repository>>,
        branch: &GitPushBranch,
    ) -> eyre::Result<()> {
        let repo = repo.lock().await;

        tracing::trace!("pulling signature from local git");
        let signature = repo.signature()?;

        tracing::trace!("fetching index and adding changed files to working tree");
        let mut index = repo.index()?;
        index.add_all(&["."], git2::IndexAddOption::DEFAULT, None)?;
        index.write()?;
        let tree = index.write_tree()?;
        let tree = repo.find_tree(tree)?;

        let parents = repo.head().map(|h| {
            h.target()
                .ok_or(eyre::anyhow!("could not fetch target"))
                .map(|t| repo.find_commit(t))
        })???;

        tracing::trace!("writing commit object");
        repo.commit(
            None,
            &signature,
            &signature,
            branch.name.to_lowercase().replace(" ", "-").as_str(),
            &tree,
            &[&parents],
        )?;

        let mut remote = repo.find_remote("origin")?;
        let head = repo.head()?;
        let refspec = &[head
            .name()
            .ok_or(eyre::anyhow!("could not find head.name"))?];

        let mut remote_callbacks = RemoteCallbacks::new();
        remote_callbacks.credentials(|url, username_from_url, _allowed_types| {
            tracing::debug!(username_from_url, url, "pulling key from ssh-agent");
            Cred::ssh_key_from_agent(username_from_url.unwrap())
        });

        let mut push_options = PushOptions::new();
        push_options.remote_callbacks(remote_callbacks);

        tracing::trace!("pushing to remote");
        remote.push(refspec, Some(&mut push_options))?;

        Ok(())
    }
}
